/* -*- mode: c++ -*- */
#ifndef __ALIBAVA_H__
#define __ALIBAVA_H__

#include <USBport.h>
#include <DataFormat.h>
#include <string>
#include <vector>
#include <glib.h>


class BitSet;
/**
 *  This class handles the USB connection. 
 *  It delegates all the operations on the port 
 *  to a port controller provided at creation time 
 * 
 * */
class Alibava
{
    public:
        typedef std::vector<DataReceiver *> ReceiverList;
        enum StateCode
        {
            None = 0,
            Reset = 0x01,
            Waiting = 0x02,
            BeetleConfig = 0x03,
            Calibration = 0x04,
            TriggerInConfig = 0x05,
            LaserSync = 0x06,
            LaserAcq = 0x07,
            LaserReading = 0x08,
            RadSourceAcq = 0x09,
            RadSourceReading = 0x0a,
            PedestalAcq = 0x0b,
            PedestalReading = 0x0c,
            AlibavaConf = 0xd,
            TriggerCounter = 0xe,
            Connect = 0x0f,
            NewData = 0x10,
            SlowControlError = 0x20,
            ADCControlError1 = 0x30,
            ADCControlError2 = 0x40,
            ADCControlError = 0x50,
            LaserAcqError = 0x60,
            SDRAMError = 0x80,
            SftReset = 0xff,
            EndOfData = 0x8000
        };
        
        enum ErrorState
        {
            NoProblem = 0,
            ReadError = -1,
            WriteError = -2,
            TimeOut = -3,
            StatusMismatch = -4,
            BoardError = -5,
            DataSizeError = -6,
            DataSizeMismatch = -7
        };


    private:
        static const int mxchan;
        bool do_stop;
        unsigned short _error;
        unsigned int   _sample_size;
        int _delay;
        int _chip_mask;
        int _nchip;
        bool _debug;
        int _fw_version;
        bool _forced_fw;
        unsigned short _fw_config;
        USBport &port; // the port controller
        ReceiverList receivers;
        //DataReceiver *_rcvr;

        GMutex *mtx;

        int beetle_header();
        int beetle_trailer();
        int trigger_counter(unsigned short val);
    public:
        /**
         * Constructor and destructor
         */
        Alibava(USBport &p);
        virtual ~Alibava();

        // Reports whether the port is in error state or not.
        bool operator!() const
        {
            return port.error()!=0;
        }

        void set_debug(bool x) { _debug=x; }

        /**
         * Methods that delegate on the port
         */

        // Finds the dev file in the different OS
        static std::string find_dev_file();

        // Opens the device
        int open(const char *fnam);

        // reconnect
        int reopen();

        // Closes the device
        int close()
        {
            return port.close();
        }

        // reads size bytes in buf
        int read(void *buf, int size)
        {
            return port.read(buf, size);
        }

        // writes size bytes from buf
        int write(void *buf, int size)
        {
            return port.write(buf, size);
        }
        
        // sets the buffer size
        void buffer_size(unsigned int sz)
        {
            port.buffer_size(sz);
        }

        // Sets a data receiver
        void set_data_receiver(DataReceiver *d)
        {
            receivers.push_back(d);
        }
        void remove_data_receiver(DataReceiver *d);

        int read_event_data(StateCode, EventData *evt);

        /** Sends a control word (8 bits)
         *  This will change the state of the machine
         */
        int send_control(unsigned char);

        /** Gets the state
         * return <0 on failure, otherwise val contains the state
         * if check != None, it will return an error if the state is not the value in check
         * If check fails, val will have in any case the value read
         * tmout is the allowed timeout. If the value is <=0, then get_state does not
         * wait for data and reads immediately
         */
        int get_state(unsigned char &val, StateCode check=None, int tmout=10000, bool notify=true);

        /* Sends a command.
         * It will change the status of the machine and will try to read
         * back the code. It calls to send_control and then get_state
         *
         * See get_state for the meaning of check and tmout
         *
         * returns <0 on failure or otherwise the status code returned by
         * the device
         */
        int send_command(StateCode cmd, StateCode check=None, int tmout=10000);

        /** Sends a data word (16 bits)
         */
        int send_word(const unsigned short &val);

        /** Check for the device asking for new data
         *  It calls to get state to know the state of the machine and
         *  returns 0 if it iw waiting for data. In case of error or the
         *  machine not ready for receiveing data, returns <0
         */
        int check_for_data();

        /** Send a data word.
         *  This one checkes for the device asking for data and if this is
         *  the case, sends the word
         */
        int send_data_word(unsigned short val);

        /**
         * Laser synchronization and beetle calibration are almost
         * identical. We do the common stuff here.
         */
        int do_calibration(StateCode code, unsigned short charge, unsigned short delay, int mxevts);

        /**
         * Starts an acquisition sequence
         */
        int do_acquisition(unsigned short, StateCode);

        /**
         *  Waits for data available
         */
        int wait_for_data(unsigned int size);
        
        /**
         * Retrieves the data after an acquisition sequence
         */
        int read_data(StateCode);

        /**
         *  Sends a reset command
         * */
        int reset(bool force=false);

        /**
         * Sends a software reset to the board
         */
        int software_reset();

        /**
         * Send the Beetle configuration
         */
        int beetle_config(int nval, short mask, unsigned short *);
        int beetle_config_comparator(int chipid, unsigned char thrs, unsigned char delta, unsigned char vrc, unsigned char icomp);
        int beetle_config_mask(int chipid, const BitSet &b);
        int beetle_config_trim(int chipid, const unsigned char *values);

        /**
         * Sets the laser delay
         */
        int set_laser_delay(int delay);
        
        /**
         * Starts calibration
         */
        int calibration(unsigned int charge, unsigned int delay, int nevts);

        /**
         * Sets the stopping flag
         */
        void stop(bool val=true)
        {
            do_stop = val;
        }

        /**
         * Returns firmware version
         */
        int firmware() const { return _fw_version; }
        void force_firmware(int x) { _fw_version=x; _forced_fw=true; }

        /**
         * returns the code of the last board code
         */
        unsigned short error() const { return _error; }
        
        /**
         * configure the system trigger
         */
        enum TriggerType
        {
            TriggerIn = 0,
            TriggerPulse=1
        };

        enum TriggerSetup
        {
            TriggerAND=0,
            TriggerOR=1
        };

        /**
         * Configures the thresholds and trigger for the RS run type.
         * Thresholds are in DAC units
         */
        int trigger(TriggerType type, TriggerSetup setup, unsigned short thrsA, unsigned short thrsB, unsigned short thrsC,
                    unsigned short thrsD);

        /**
         * Laser Synchronization
         */
        int laser_sync(unsigned short delay, int nevts);

        /**
         * Laser Acquisition
         */
        int laser_acq(unsigned short nevts);

        /**
         * Laser Reading
         */
        int laser_reading();

        /**
         * Rad Sourse Acquisition
         */
        int rad_source_acq(unsigned short nevts);

        /**
         * Rad Sourse Reading
         */
        int rad_source_reading();
        
        /**
         * Pedestal acquisition
         */
        int pedestal_acq(unsigned short nevts);
        
        /**
         * Pedestal Reading
         */
        int pedestal_reading();
        

        /*
         * Trigger counter
         */
        int get_trigger_counter();
        void reset_trigger_counter();

        /*
         * Configure Firmware
         */
        enum ConfigFirmware
        {
            EnableBusy = 1,
            PulseShape = 2
        };
        int config_firmware(unsigned short val);
        unsigned short get_firmware_config() const { return _fw_config; }

        int delay() const 
        {
            return _delay;
        }
        void delay(int x)
        {
            _delay = x;
        }
        
        int mask() const { return _chip_mask; }
        void mask(int m) { _chip_mask = m; _nchip = ( m == 3 ? 2 : 1 );}

        int nchip() const { return _nchip; }

        bool try_lock() const;
        void unlock();
        void lock();

        std::string error_string(int) const;

        /*
         * Functions related to the DataReceivers
         */
        void define_scan(int npts, double from, double to);
        void end_of_scan_point(double value, unsigned int ntrg);

};

#endif /*ALIBAVA_H_*/
