#include <glib.h>
#ifdef HAVE_CONFIG_H
#ifdef G_OS_WIN32
#include "config_win.h"
#else
#include <config.h>
#endif
#endif

#include <iostream>


#ifdef G_OS_WIN32
#include <windows.h>
#else
#include <d2xx/WinTypes.h>
#endif


#ifdef HAS_FTD2XX
#include <d2xx/ftd2xx.h>
#else
typedef DWORD           *FT_HANDLE;
typedef ULONG            FT_STATUS;
#endif

#include "USBd2xx.h"


struct USBd2xx_priv
{
    FT_HANDLE fthandle;
    FT_STATUS ftstatus;
    DWORD numdevs;
    DWORD flags;
    DWORD locID;
    DWORD deviceID;
    DWORD type;
    DWORD ftdevice;
    char serialnumber[16];
    char description[64];
    DWORD deviceindex;

    GMutex *eMutex;
    GCond *eConVar;
    int iVar;
};

USBd2xx::USBd2xx()
{
    init_mutex();
}

USBd2xx::USBd2xx(const char *fnam) :
    USBport(fnam)
{
    init_mutex();
    open(fnam);
}

USBd2xx::~USBd2xx()
{
    g_mutex_free(_priv->eMutex);
    g_cond_free(_priv->eConVar);
    g_free( _priv );
}

void USBd2xx::init_mutex()
{
    std::cout << "Loading USBd2xx" << std::endl;
    _priv = g_new0(USBd2xx_priv, 1);
    _priv->fthandle = 0;
    _priv->ftstatus = 0;
    _priv->eMutex = g_mutex_new();
    _priv->eConVar = g_cond_new();
    std::cout << "...done" << std::endl;
    if (!connected())
    {
        std::cout << "It seems that the device is not connected" << std::endl;
    }
}


int USBd2xx::error() const
{
    return _priv->ftstatus;
}

int USBd2xx::fildes() const
{
    return _priv->deviceindex;
}

#ifdef HAS_FTD2XX
bool USBd2xx::connected () {
    _priv->ftstatus = FT_CreateDeviceInfoList(&_priv->numdevs);
    if (_priv->ftstatus != FT_OK) {
        return false;
    }
    if (_priv->numdevs > 0 ) {
        for (DWORD i = 0; i < _priv->numdevs; i++) {
            _priv->ftstatus = FT_GetDeviceInfoDetail(i,&_priv->flags,&_priv->type,&_priv->deviceID,&_priv->locID,_priv->serialnumber,_priv->description,&_priv->fthandle);
            if (_priv->ftstatus == FT_OK) {
                if (_priv->deviceID == 0x04036001) {
                    _priv->deviceindex = i;
                    break;
                }
            }
            else {
                break;
            }
        }
        if (_priv->ftstatus == FT_OK) {
            return true;
        }
        else {
            return false;
        }
    }
    else {
        return false;    
    }
}
int USBd2xx::initialize(unsigned int baudrate, unsigned int rdtmout, unsigned int wrtmout)
{
    unsigned int br = baudrate, rd=rdtmout, wr=wrtmout;
    _priv->ftstatus = FT_SetDataCharacteristics(_priv->fthandle,
                                                FT_BITS_8,
                                                FT_STOP_BITS_1,
                                                FT_PARITY_NONE);
    if (_priv->ftstatus != FT_OK)
    {
        return -1;
    }
    _priv->ftstatus = FT_SetBaudRate(_priv->fthandle, br);
    if (_priv->ftstatus != FT_OK)
    {
        return -1;
    }
    _priv->ftstatus = FT_SetTimeouts(_priv->fthandle, rd, wr);
    if (_priv->ftstatus != FT_OK)
    {
        return -1;
    }
    _priv->ftstatus = FT_SetLatencyTimer(_priv->fthandle, 2);
    if (_priv->ftstatus != FT_OK) {
    	return -1;
    }
    return 0;
}
int USBd2xx::open(const std::string &fnam)
{
    if (_priv->numdevs > 0)
    {
        _priv->ftstatus = FT_Open(_priv->deviceindex, &_priv->fthandle);
        if (_priv->ftstatus == FT_OK)
        {
            _priv->ftstatus = FT_GetDeviceInfo(_priv->fthandle,
                                               &_priv->ftdevice,
                                               &_priv->deviceID,
                                               _priv->serialnumber,
                                               _priv->description,
                                               NULL);
            if (initialize(115200,5000,1000))
            {
                return -1;
            }

            return _priv->deviceindex;
        }
        else
        {
            return -1;
        }
    }
    else
    {
        return -1;
    }
}

int USBd2xx::close()
{
    if (_priv->numdevs > 0)
    {
        _priv->ftstatus = FT_Close(_priv->fthandle);
        return  (_priv->ftstatus == FT_OK ? 0 : -1 );
    }
    else
    {
        return -1;
    }
}

int USBd2xx::read(void *buf, int size)
{
    DWORD br;
    _priv->ftstatus = FT_Read(_priv->fthandle,(unsigned char *)buf,size,&br);
    if (_priv->ftstatus != FT_OK)
    {
        return -1;
    }
    return int(br);
}

int USBd2xx::write(void *buf, int size)
{
    DWORD bw;
    _priv->ftstatus = FT_Write(_priv->fthandle, (unsigned char *)buf, size, &bw);
    if (_priv->ftstatus != FT_OK)
    {
        return -1;
    }
    return int(bw);
}

unsigned int USBd2xx::data_available(unsigned int size)
{
    DWORD rxb;
    _priv->ftstatus = FT_GetQueueStatus(_priv->fthandle, &rxb);
    if (_priv->ftstatus != FT_OK)
    {
        return 0;
    }
    return (unsigned int)(rxb);
}

int USBd2xx::wait_for_data(double tmout)
{
    int rc = poll(tmout);
    if (rc)
        return 0;
    
    DWORD rxb;
    _priv->ftstatus = FT_GetQueueStatus(_priv->fthandle, &rxb);
    return  _priv->ftstatus != FT_OK ? 0 : rxb;
}
int USBd2xx::poll(int timeout)
{
#ifdef G_OS_WIN32
#ifdef __does_not_work__
	// First, create the event and call FT_SetEventNotification.
	FT_HANDLE ftHandle; // handle of an open device
	FT_STATUS ftStatus;
	HANDLE hEvent;
	DWORD EventMask;
	hEvent = CreateEvent(
			NULL,
			false, // auto-reset event
			false, // non-signalled state
			""
	);
	EventMask = FT_EVENT_RXCHAR;
	ftStatus = FT_SetEventNotification(_priv->fthandle,EventMask,hEvent);

	WaitForSingleObject(hEvent,timeout);

	DWORD EventDWord=0;
	DWORD RxBytes=0;
	DWORD TxBytes=0;
	FT_GetStatus(ftHandle,&RxBytes,&TxBytes,&EventDWord);
	if (RxBytes > 0) {
		return 0;
	}
	else
		return 1;
#else
	return 0;
#endif
#else
    _priv->ftstatus = FT_SetEventNotification(_priv->fthandle, FT_EVENT_RXCHAR, (PVOID)this);
    if( _priv->ftstatus != FT_OK )
    {
        return -1;
    }

    int rc;
    GTimeVal ts;
    g_get_current_time(&ts);

    ts.tv_usec += (timeout % 1000);
    ts.tv_sec += (timeout + ts.tv_usec) / 1000;
    ts.tv_usec = ts.tv_usec % 1000;

    g_mutex_lock(_priv->eMutex);
    rc = g_cond_timed_wait(_priv->eConVar, _priv->eMutex, &ts);
    g_mutex_unlock(_priv->eMutex);
    if (rc)
    {
        return 1;
    }

    return 0;
#endif
}
#else
bool USBd2xx::connected () {
    return false;
}
int USBd2xx::initialize(unsigned int baudrate, unsigned int rdtmout, unsigned int wrtmout)
{
    return 0;
}
int USBd2xx::open(const std::string &fnam)
{
    return 0;
}
int USBd2xx::close()
{
    return 0;
}
int USBd2xx::read(void *buf, int size)
{
    return 0;
}
int USBd2xx::write(void *buf, int size)
{
    return 0;
}
unsigned int USBd2xx::data_available(unsigned int size)
{
    return 0;
}
int USBd2xx::wait_for_data(double tmout)
{
    return 0;
}
int USBd2xx::poll(int timeout)
{
    return 0;
}
#endif
